/*
 Copyright 2005 Kapor Enterprises, Inc. and Todd Agulnick.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 foxmarks-bookmarkset.js: implments class BookmarkSet, which wraps an RDF
 datasource containing bookmark data. Some of this code deals with the
 complexity introduced by the bookmark manager itself wrapping an RDF datasource.

 */

function BookmarkSet()
{
  this._init();
}

BookmarkSet.prototype =
{
  ds: null,
  rdfs: null,
  rdfcu: null,
  container: null,
  dirtyList: null,
  lastModifiedPredicate: null,
  addDatePredicate: null,
  typePredicate: null,
  timer: null,

  _init: function()
  {
    this.rdfs = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(
      Components.interfaces.nsIRDFService);
    this.container = Components.classes["@mozilla.org/rdf/container;1"].
      createInstance(Components.interfaces.nsIRDFContainer);
    this.rdfcu = Components.classes["@mozilla.org/rdf/container-utils;1"].
      getService(Components.interfaces.nsIRDFContainerUtils);
    this.lastModifiedPredicate = this.rdfs.
      GetResource("http://home.netscape.com/WEB-rdf#LastModifiedDate");
    this.addDatePredicate = this.rdfs.
      GetResource("http://home.netscape.com/NC-rdf#BookmarkAddDate");
    this.typePredicate = this.rdfs.
      GetResource("http://www.w3.org/1999/02/22-rdf-syntax-ns#type");

  },

  // traverses this's bookmarks hierarchy starting with
  // startnode, calling action(node) for each node in the tree,
  // then calling complete() when traversal is done.
  // enforces rules about maximum run times to prevent hanging the UI
  // when traversing large trees or when running on slow CPU's.
  // action() should return true to continue, false to abort.
  // complete() is called with true if traversal completed, false if aborted
  OnTree: function(startnode, action, complete)
  {
    function ElapsedMS(s) {
      var now = new Date();
      return now.getTime() - s.getTime();
    }

    var Continue = {
      parent: this,
      notify: function(timer) {

        var set = this.parent;
        var result;
        var s = new Date();
        while (resources.length > 0 && ElapsedMS(s) < 100) {
          var resource = resources.shift();

          result = action(set, resource);

          if (!result)
            break;

          if (set.IsContainer(resource)) {
            set.container.Init(set.ds, resource);

            var children = set.container.GetElements();
            while (children.hasMoreElements()) {
              resources.push(children.getNext());
            }
          }
        }

        if (resources.length > 0 && result) {
          timer.initWithCallback(Continue, 10,
            Components.interfaces.nsITimer.TYPE_ONE_SHOT);
        } else {
          complete(set, result);
        }
      }
    }

    if (!startnode) {
      startnode = this.root;
    }

    var resources = [startnode];

    this.timer = Components.classes["@mozilla.org/timer;1"].
      createInstance(Components.interfaces.nsITimer);

    this.timer.initWithCallback(Continue, 10,
      Components.interfaces.nsITimer.TYPE_ONE_SHOT);

    return;

  },

  SetDatasource: function(uri)
  {
    this.ds = this.rdfs.GetDataSource(uri);
    this.ds.QueryInterface(Components.interfaces.nsIBookmarksService);
  },

  CreateEmptyDatasource: function()
  {
    this.ds = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
      .createInstance(Components.interfaces.nsIRDFDataSource);
  },

  DatasourceIsValid: function()
  {
    var resources = [this.root];

    while (resources.length > 0) {
      var resource = resources.shift();

      if (!this.ResourceExists(resource))
         return false;

      // this a folder? If so, get its children and add them to the list for examination
      if (this.IsContainer(resource)) {
        this.container.Init(this.ds, resource);

        var children = this.container.GetElements();
        while (children.hasMoreElements()) {
          resources.push(children.getNext().QueryInterface(Components.interfaces.nsIRDFResource));
        }
      }
    }

    return true;
  },

  CreateFromString: function(str, attrs)
  {
    // random base URI for parseFromString function
    var url = "http://www.foxcloud.com/";

    this.CreateEmptyDatasource();

    var parser = Components.classes["@mozilla.org/rdf/xml-parser;1"]
      .createInstance(Components.interfaces.nsIRDFXMLParser);

    var ios = Components.classes["@mozilla.org/network/io-service;1"]
      .getService(Components.interfaces.nsIIOService);
    var baseUri = ios.newURI(url,null,null);

    parser.parseString(this.ds, baseUri, str);

    if (!this.DatasourceIsValid()) {
      LogWrite("Parser error; file listed below >>>\n" + str + "\n<<<");
      return -1;
    }

    // parse out version and token
    if (attrs) {
      var version = str.match(/Foxmarks\/([0-9.]+)/);
      var token = str.match(/Token\/(\w+)/);

      if (version && version.length > 1) {
        attrs.version = version[1];
      } else {
        attrs.version = "";
      }
      
      if (token && token.length > 1) {
        attrs.token = token[1];
      } else {
        attrs.token = "";
      }
    }

    return 0;
  },

  SerializeToString: function(token)
  {
    var s = new Serializer(this.ds);
    var string = s.Serialize();
    var header = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
        "<!-- Foxmarks/" + FoxmarksVersion() + " Token/" + token + " -->\n";

    return header + string;
  },

  get root()
  {
    return this.rdfs.GetResource("NC:BookmarksTopRoot");
  },

  // remove all assertions associated with a resource
  RemoveResource: function(resource)
  {
    var predicates = this.ds.ArcLabelsOut(resource);

    while (predicates.hasMoreElements()) {
      var predicate = predicates.getNext();
      var target = this.ds.GetTarget(resource, predicate, true);
      this.ds.Unassert(resource, predicate, target);

      // if we remove URL, we also lose forwardProxy; this
      // causes our enumerator to reference a no-longer existing object
      // and we crash. To avoid this problem, re-calculate remaining predicates
      if (predicate instanceof Components.interfaces.nsIRDFResource &&
          predicate.Value == "http://home.netscape.com/NC-rdf#URL") {
        predicates = this.ds.ArcLabelsOut(resource);
      }
    }
  },

  // check whether two resources are (for our purposes) equivalent
  IsEquivalent: function(source, resource)
  {
    return (this.HasAllAssertions(source, resource) && source.HasAllAssertions(this, resource));
  },

  // returns true if source:resource has all assertions that we have
  HasAllAssertions: function(source, resource)
  {
    // enumerate our assertions
    var alo = this.ds.ArcLabelsOut(resource);

    while(alo.hasMoreElements()) {
      var predicate = alo.getNext();
      var target = this.ds.GetTarget(resource, predicate, true);
      var otherTarget = source.ds.GetTarget(resource, predicate, true);

      // don't bother with null targets
      if (!target)
        continue;


      predicate.QueryInterface(Components.interfaces.nsIRDFResource);

      // ignore differences in last modified date
      if (predicate.Value == "http://home.netscape.com/WEB-rdf#LastModifiedDate")
        continue;

      if (predicate.Value == "http://home.netscape.com/WEB-rdf#LastVisitDate")
         continue;

      if (predicate.Value == "http://home.netscape.com/WEB-rdf#LastCharset")
        continue;

      if (predicate.Value == "http://home.netscape.com/NC-rdf#Icon")
        continue;

      if (predicate.Value == "http://home.netscape.com/NC-rdf#ID")
        continue;

      if (predicate.Value == "http://developer.mozilla.org/rdf/vocabulary/forward-proxy#forward-proxy")
        continue;

      if (predicate.Value == "http://www.w3.org/1999/02/22-rdf-syntax-ns#type" &&
         target instanceof Components.interfaces.nsIRDFResource &&
         (target.Value == "http://home.netscape.com/NC-rdf#Folder" ||
         target.Value == "http://home.netscape.com/NC-rdf#Bookmark"))
        continue;

      if (!source.ds.HasAssertion(resource, predicate, target, true)) {
/*
        LogWrite("Resource " + this.ResourceName(resource) + " is missing " + predicate.Value);
        if (target instanceof Components.interfaces.nsIRDFResource)
          LogWrite("We have " + target.Value);
        else
          LogWrite("We have a " + target);
        if (otherTarget instanceof Components.interfaces.nsIRDFResource)
          LogWrite( "he's got " + otherTarget.Value);
        else
          LogWrite("he's got a " + otherTarget);
*/
        return false;
      }
    }

    return true;
  },

  // copy a resource and its statements from another Bookmarkset
  CopyResource: function(source, resource, logit)
  {
    var type = source.ds.GetTarget(resource, this.typePredicate, true);
    var BMSVC = this.ds;
    var isBMSVC = (BMSVC instanceof Components.interfaces.nsIBookmarksService);

    // don't bother copying if it's already there
    if (this.ResourceExists(resource) && this.IsEquivalent(source, resource)) {
      return;
    }

    if (logit)
      LogWrite("Copying dirty resource: " + source.ResourceName(resource));

    if (!(type instanceof Components.interfaces.nsIRDFResource)) {
      resource.QueryInterface(Components.interfaces.nsIRDFResource);
      LogWrite("WARNING: CopyResource: Found no type associated with\n" + resource.Value);
      return;
    }

    // we'd like to remove all resources associated with the
    // destination resource, but awful things happen if we try
    // to do so on a Livemark. Curse you, Livemarks!
    // XXX note that changes to a livemark's name (or other properties
    // that might be safe to change) aren't deleted either,
    // so those changes might not be synchronized properly
    if (type.Value != "http://home.netscape.com/NC-rdf#Livemark") {
      this.RemoveResource(resource);
    } else {
      // dump("skipping resource remove for livemark " + resource.Value);
    }

    // note that for folders and bookmarks, the type predicate is not explicitly set
    // but rather synthesized by bookmarks service code based on whether the resource
    // has an instanceOf tag. We assert explicitly because type doesn't show up in ArcLabelsOut.
    this.ds.Assert(resource, this.typePredicate, type, true);

    var predicates = source.ds.ArcLabelsOut(resource);

    while (predicates.hasMoreElements()) {
      var predicate = predicates.getNext();
      var target = source.ds.GetTarget(resource, predicate, true);

      // sometimes, targets don't exist (icons in particular)
      // just skip 'em
      if (!target)
        continue;

      predicate.QueryInterface(Components.interfaces.nsIRDFResource);

      // don't assert last mod on bmsvc
      if (predicate.Value == "http://home.netscape.com/WEB-rdf#LastModifiedDate" && isBMSVC)
        continue;

      try {
        this.ds.Assert(resource, predicate, target, true);
      } catch (e) {
        LogWrite("ERROR: assert failed copying " + source.ResourceName(resource));
      }

      target.QueryInterface(Components.interfaces.nsIRDFNode);
      if (!this.ds.HasAssertion(resource, predicate, target, true)) {

        // an assertion we made didn't actually stick in the RDF data store.
        // Handle a few special cases, otherwise gripe.

        if (predicate.Value == "http://home.netscape.com/WEB-rdf#LastCharset") {
          continue;
        }

        if (resource.Value == "NC:BookmarksTopRoot") {
          continue;
        }

        // Hack: if we were refused on BookmarksToolbarFolder, use the bookmarks service
        // routine to explicitly set the bookmarks toolbar folder to this resource
        if (predicate.Value == "http://home.netscape.com/NC-rdf#BookmarksToolbarFolder") {
          if (isBMSVC) {
            this.ds.endUpdateBatch(); // force the bookmarkstoolbarfolder to update
            BMSVC.setBookmarksToolbarFolder(resource);
            this.ds.beginUpdateBatch();
            continue;
          } else {
            alert("Tried and failed to set ToolbarFolder on " + resource.Value);
            continue;
          }
        }

        LogWrite("WARNING: Assertion didn't stick: <" + resource.Value + "," + predicate.Value + "," + target.Value + ">");

      }

    }
  },

  IsContainer: function(resource)
  {
    if (!this.rdfcu.IsContainer(this.ds, resource))
      return false;

    // if it's a Livemark, don't consider it a container, as we don't process its children
    var type = this.ds.GetTarget(resource, this.typePredicate, true);

    if (type instanceof Components.interfaces.nsIRDFResource) {
      if (type.Value == "http://home.netscape.com/NC-rdf#Livemark") {
        return false;
      }
    }

    return true;
  },
  
  GetType: function(resource)
  {
    var type = this.ds.GetTarget(resource, this.typePredicate, true)
    
    if (type instanceof Components.interfaces.nsIRDFResource) {
      return type.Value;
    } else {
      return null;
    }
  },
    

  IsSeparator: function(resource)
  {
    return (this.GetType(resource) == "http://home.netscape.com/NC-rdf#BookmarkSeparator");
  },

  IsLivemark: function(resource)
  {
    return (this.GetType(resource) == "http://home.netscape.com/NC-rdf#Livemark");
  },
  
  IsMicsum: function(resource)
  {
    return (this.GetType(resource) == "http://home.netscape.com/NC-rdf#MicsumBookmark");
  },



  // copy a subtree from another set into us
  // copy entire tree if subtree is null
  // markDirty:
  //  false       don't mark anything dirty
  //  true        mark all resources dirty depending on baseDate
  //  baseDate    if markDirty is true, mark resources dirty that are dirty relative
  //              to baseDate. If baseDate is null, mark all resources dirty

  CopyFromDatasourceAS: function(source, subtree, markDirty, baseDate, callback)
  {
    if (!this.ds) {
      this.CreateEmptyDatasource();
    }

    var base = 0;

    if (baseDate)
      base = Date.parse(baseDate) * 1000 + 1000000;

    this.ds.beginUpdateBatch();

    this.OnTree(subtree, Action, Complete);
    return;

    function Action(set, resource) {

      // copy all statements associated with this resource
      set.CopyResource(source, resource, false);
      if (markDirty && (!base || set.IsDirty(resource, base))) {
        LogWrite("CFD: Marking resource dirty: " + set.ResourceName(resource));
        set.MarkDirty(resource);
      }

      return true;
    }

    function Complete(set, okay) {
      set.ds.endUpdateBatch();
      callback();
    }
  },

  // returns null if resource isn't dirty; otherwise, returns
  // the dirty date value (from either last mod or add date)
  IsDirty: function(resource, baseDate)
  {
    var lastModifiedDate;
    var primary;
    var backup;
    var isLivemarkOrMicsum;

    // if it's a livemark, determine dirty state based on add date.
    // if add date is absent (which happens for the livemark
    // included in the default set), it's not dirty. C'est la vie.
    // for all others, use last modified date, with add date as a backup
    isLivemarkOrMicsum = this.IsLivemark(resource) || this.IsMicsum(resource);
    if (isLivemarkOrMicsum) {
      primary = this.addDatePredicate;
      backup = this.addDatePredicate;
    } else {
      primary = this.lastModifiedPredicate;
      backup = this.addDatePredicate;
    }

    var target = this.ds.GetTarget(resource, primary, true);
    if (target instanceof Components.interfaces.nsIRDFDate) {
      lastModifiedDate = target.Value;
    } else {  // missed primary; try backup
      target = this.ds.GetTarget(resource, backup, true);
      if (target instanceof Components.interfaces.nsIRDFDate) {
        lastModifiedDate = target.Value;
      } else {  // no date; assume it's not dirty
        return null;
      }
    }

    return lastModifiedDate > baseDate ? lastModifiedDate : null;
  },


  IsNew: function(resource, baseDate)
  {
   var addDate;

   var target = this.ds.GetTarget(resource, this.addDatePredicate, true);
   if (target instanceof Components.interfaces.nsIRDFDate)
      return target.Value > baseDate;
   else
      return false;
  },

  // passes back true if any dirty resources were discovered
  BuildDirtyListAS: function(lastSyncDate, callback)
  {
    this.dirtyList = new Array(0);

    this.OnTree(null, Action, Complete);
    return;

    function Action(set, resource) {

      // item was modified since our last sync; add to dirty list
      if (set.IsDirty(resource, lastSyncDate)) {
        LogWrite("Dirty: " + set.ResourceName(resource));
        set.dirtyList.push(resource);
      }

      return true;
    }

    function Complete(set, okay) {
      callback(set.dirtyList.length > 0);
    }

  },

  // returns true if any dirty resources were discovered
  HasDirtyResourcesAS: function(base, callback)
  {
    // convert lastSync string to format used by nsIRDFDate (microseconds since 1/1/1970)
    // ... and round up by 1 second since that is the limit of accuracy on the recorded date
    var lastSyncDate = Date.parse(base) * 1000 + 1000000;
    var now = new Date();
    var maxFutureDate = now.getTime() * 1000 + 1000000;

    this.OnTree(null, Action, Complete);
    return;

    function Action(set, resource) {

      // item was modified since our last sync
      var dirtyDate = set.IsDirty(resource, lastSyncDate);
      
      if (dirtyDate) {
        // item's date more recent than the lastSyncDate;
        // but before we call it dirty (and thus trigger sync), let's
        // check to see if it's some whacko in-the-future date.
        if (dirtyDate > maxFutureDate) {
          LogWrite("WARNING: Future date for " + set.ResourceName(resource));
          return true;   // whacko date
        }
        return false;      // dirty, not whacko
      }
      return true;         // not dirty
    }

    // abnormal termination in this case means
    // that we found something dirty
    function Complete(set, okay) {
      callback(!okay);
    }
  },

  // mark as dirty again any resources that appear dirty
  // (useful if we marked something dirty before, but then
  // spent a lot of time in UI. This causes the last-modified
  // attribute to set to NOW.
  RedirtyResourcesAS: function(base, callback)
  {
    this.OnTree(null, Action, Complete);
    return;

    function Action(set, resource) {

      // item was modified during sync, mark it again
      if (set.IsDirty(resource, base)) {
        set.MarkDirty(resource);
      }

      return true;
    }

    function Complete(set, okay) {
      callback();
    }
  },

  // for each item inside of source's resource (a folder),
  // check to see whether the item's parent (within us) is in our dirty list.
  // Returns true if some element is found.
  ConflictedParent: function(source, resource)
  {
    source.container.Init(source.ds, resource);

    var children = source.container.GetElements();
    while (children.hasMoreElements()) {
      var child = children.getNext();
      var parent = this.GetParent(child);

      if (parent && this.IndexInDirtyList(parent) >= 0)
        return true;
    }

    return false;
  },


  RemoveChildrenFromParents: function(source, resource)
  {
    // we're about to copy resource from source to this.
    // examine each of resource's children, and remove them
    // from their parent in us.

    this.container.Init(source.ds, resource);

    var children = this.container.GetElements();
    while (children.hasMoreElements()) {
      var child = children.getNext();
      var parent;

      while (parent = this.GetParent(child)) {

        var container = Components.classes["@mozilla.org/rdf/container;1"].
            createInstance(Components.interfaces.nsIRDFContainer);

        container.Init(this.ds, parent);
        container.RemoveElement(child, true);
        this.MarkDirty(parent);
      }
    }
  },

  IndexInDirtyList: function(resource)
  {
    var uri = resource.Value;

    for (var index = 0; index < this.dirtyList.length; ++index) {
      if (this.dirtyList[index].EqualsString(uri))
        return index;
    }

    return -1;
  },

  ProcessDirtyList: function(source, lastSyncDate, toBeResolved)
  {
    var dlindex;

    for (dlindex in source.dirtyList) {
      var resource = source.dirtyList[dlindex];

      var index = this.IndexInDirtyList(resource);
      if (index >= 0) {
        LogWrite("Conflicted resource: " + this.ResourceName(resource));
        toBeResolved.push(resource);  // add to list to be resolved
        continue;
      }

      if (source.IsContainer(resource)) {

        // we're about to copy a folder. For each child of resource's children,
        // check to see whether its parent on our side is dirty. If it is, we have
        // a potential parent conflict, so instead of doing the copy, add resource
        // to the conflict list

        if (this.ConflictedParent(source, resource)) {
          LogWrite("Conflicted parent " + source.ResourceName(resource));
          toBeResolved.push(resource);
          if (!this.ResourceExists(resource)) {
            LogWrite("WARNING: Conflicted parent is new; copying empty folder");
            this.CopyFolderWithoutChildren(source, resource);
          }
          continue;
        }

        // before copying a folder, remove folder's children from any other parents
        this.RemoveChildrenFromParents(source, resource);
      }

      this.CopyResource(source, resource, true);
      this.MarkDirty(resource);

    }

    for (resource in toBeResolved) {
      index = this.IndexInDirtyList(toBeResolved[resource]);
      if (index >= 0)
        this.dirtyList.splice(index, 1);    // remove it from destDirty so we don't process it again
    }
  },

  // Because separators have ephemeral id's, they occasionally
  // get orphaned during the sync process. This routine finds them
  // and turns them back into separators for a single folder.
  FixSeparators: function(resource)
  {
    var typeTarget = this.rdfs.GetResource("http://home.netscape.com/NC-rdf#BookmarkSeparator");

    if (this.IsContainer(resource)) {
      this.container.Init(this.ds, resource);

      var children = this.container.GetElements();
      while (children.hasMoreElements()) {
        var child = children.getNext();
        var alo = this.ds.ArcLabelsOut(child);

        // We found a node with nothing attached.
        // assume it's an orphaned separator and assert its type
        if (!alo.hasMoreElements()) {
          this.ds.Assert(child, this.typePredicate, typeTarget, true);
        }
      }
    }
  },

  // This is a post-processing pass that repairs any dangling references
  // that we may have inadvertently created. Many of these will be separators,
  // with their ephemeral id's. But others will be first class resources
  // that somehow got overlooked.
  FixDanglingReferences: function(source, callback)
  {
    var typeTarget = this.rdfs.
        GetResource("http://home.netscape.com/NC-rdf#BookmarkSeparator");

    this.OnTree(null, Action, Complete);
    return;

    function Action(set, resource) {

      var alo = set.ds.ArcLabelsOut(resource);

      // We found a node with nothing attached.

      if (!alo.hasMoreElements()) {

        // somehow, resource didn't get copied properly during
        // regular sync process. So copy it now.

        if (source.ResourceExists(resource)) {
          if (source.IsContainer(resource)) {
            source.RemoveChildrenFromParents(source, resource);
          }

          set.CopyResource(source, resource, true);
          set.MarkDirty(resource);
        } else {
          // assume it's an orphaned separator and assert its type
          set.ds.Assert(resource, set.typePredicate, typeTarget, true);
        }

        LogWrite("WARNING: Transformed dangling reference into " +
          set.ResourceName(resource));
      }

      return true;
    }

    function Complete(set, okay) {
      callback();
    }
  },

  Prop: function(resource, propname)
  {
    var predicate = this.rdfs.GetResource(propname);
    return this.ds.GetTarget(resource, predicate, true);
  },

  PropString: function(resource, propname)
  {
    var rv = this.Prop(resource, propname);
    if (rv instanceof Components.interfaces.nsIRDFLiteral) {
      return rv.Value;
    } else {
      return "";
    }
  },

  CopyProperty: function(resource, propname, srcset)
  {
    var oldProp = this.Prop(resource, propname);
    var newProp = srcset.Prop(resource, propname);

    if (oldProp) {
      this.ds.Unassert(resource, this.rdfs.GetResource(propname), oldProp);
    }

    if (newProp) {
      this.ds.Assert(resource, this.rdfs.GetResource(propname), newProp, true);
    }
  },

  // copy resource's child references from srcset to us
  CopyChildren: function(resource, srcset)
  {
    this.container.Init(this.ds, resource);
    srcset.container.Init(srcset.ds, resource);

    var count = this.container.GetCount();

    while (count > 0) {
      this.container.RemoveElementAt(count, true);
      count--;
    }

    var children = srcset.container.GetElements();

    while (children.hasMoreElements()) {
      var child = children.getNext();

      this.container.AppendElement(child);
    }

    return;
  },

  // return the parent of this resource (null if not found)
  GetParent: function(resource)
  {

    var parent = null;

    var arcsIn = this.ds.ArcLabelsIn(resource);

    while (arcsIn.hasMoreElements()){
      var arc = arcsIn.getNext();

      if (arc instanceof Components.interfaces.nsIRDFResource) {
        if (this.rdfcu.IsOrdinalProperty(arc)) {
          parent = this.ds.GetSource(arc, resource, true);
          break;
        }
      }
    }

    return parent;
  },

  // determine whether resource exists
  ResourceExists: function(resource)
  {
    var alo = this.ds.ArcLabelsOut(resource);
    return alo.hasMoreElements();
  },

  // make sure that this resource has its last-modified field set appropriately
  MarkDirty: function(resource)
  {
    var BMSVC = this.ds;
    var pred = (this.IsLivemark(resource) || this.IsMicsum(resource)) ? this.addDatePredicate : this.lastModifiedPredicate;

    // never assert last modified date on bookmarks service; causes
    // all sorts of trouble because we wind up with multiple last mod
    // date asssertions. Just let bookmarks service handle it for us.
    if (BMSVC instanceof Components.interfaces.nsIBookmarksService)
      return;

    var lastModDate = this.rdfs.GetDateLiteral(Date.parse(Date()) * 1000);
    var oldModDate = this.ds.GetTarget(resource, pred, true);

    if (oldModDate) {
      this.ds.Unassert(resource, pred, oldModDate, true);
    }

    this.ds.Assert(resource, pred, lastModDate, true);
  },

  MarkNew: function(resource)
  {
    var addDate = this.rdfs.GetDateLiteral(Date.parse(Date()) * 1000);
    var oldDate = this.ds.GetTarget(resource, this.addDatePredicate, true);

    if (oldDate) {
      this.ds.Unassert(resource, this.addDatePredicate, oldDate, true);
    }

    this.ds.Assert(resource, this.addDatePredicate, addDate, true);
  },

  ResourceName: function(resource)
  {
    const propName = "http://home.netscape.com/NC-rdf#Name";
    const propType = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type";

    var rname = this.PropString(resource, propName);
    var rtype = this.Prop(resource, propType);

    if (rtype instanceof Components.interfaces.nsIRDFResource) {
      rtype = rtype.Value;

      if (rtype.indexOf("#") > 0)
        rtype = rtype.substr(rtype.indexOf("#") + 1);
    } else {
      rtype = "<notype>";
    }

    if (!rname.length)
      rname = "<noname>";

    return rtype + ":" + rname + " {" + resource.QueryInterface(Components.interfaces.nsIRDFResource).Value + "}";
  },

  EnumerateChildren: function(resource)
  {
    var result = new Array(0);

    if (!this.IsContainer(resource))
      return result;

    this.container.Init(this.ds, resource);

    var children = this.container.GetElements();
    while (children.hasMoreElements()) {
      result.push(children.getNext().QueryInterface(Components.interfaces.nsIRDFResource));
    }

    return result;
  },

  // returns resource id of created separator
  // client must insert the separator in the appropriate parent folder
  // if name is non-null, it will be set as the separator's name
  CreateSeparator: function(name)
  {
    const litSep = "http://home.netscape.com/NC-rdf#BookmarkSeparator";
    var resource = this.rdfs.GetAnonymousResource();
    var resSep = this.rdfs.GetResource(litSep);
    this.ds.Assert(resource, this.typePredicate, resSep, true);
    if (name) {
      const litName = "http://home.netscape.com/NC-rdf#Name";
      var resName = this.rdfs.GetResource(litName);
      var resValue = this.rdfs.GetLiteral(name);
      this.ds.Assert(resource, resName, resValue, true);
    }
    this.MarkDirty(resource);
    return resource;
  },

  CopyResourceAnonymous: function(source, resource)
  {
    if (!source.ResourceExists(resource))
       return null;

    var id = this.rdfs.GetAnonymousResource();
    var alo = source.ds.ArcLabelsOut(resource);
    var isContainer = source.IsContainer(resource);

    while (alo.hasMoreElements()) {
      var pred = alo.getNext();
      var target = source.ds.GetTarget(resource, pred, true);
      if (isContainer && source.rdfcu.IsOrdinalProperty(pred)) {
         target = this.CopyResourceAnonymous(source, target);
      }
      if (target) {
         this.ds.Assert(id, pred, target, true);
      }
    }

    this.MarkDirty(id);

    LogWrite("Created new resource" + this.ResourceName(id));

    return id;
  },
  
  CopyFolderWithoutChildren: function(source, resource)
  {
    if (!source.ResourceExists(resource))
      return;
      
    var alo = source.ds.ArcLabelsOut(resource);
    
    while (alo.hasMoreElements()) {
      var pred = alo.getNext();
      var target = source.ds.GetTarget(resource, pred, true);
      if (!source.rdfcu.IsOrdinalProperty(pred)) {
        this.ds.Assert(resource, pred, target, true);
      }
    }
    
    this.MarkDirty(resource);
  }
}
