/* 
 Copyright 2005 Kapor Enterprises, Inc. and Todd Agulnick.
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 	http://www.apache.org/licenses/LICENSE-2.0
 	
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
 foxmarks-sync.js: implements the core synchronization algorithm, including
 the logic for conflict resolution.
  
 */

// Sync two bookmarksets

// return val:
// -1: sync failed -- unresolved conflict or user canceled
//  0: sync succeeded
//  1: sync succeeded, no update required

function Sync(local, remote, lastSync, dirty, callback)
{
	// convert lastSync string to format used by nsIRDFDate (microseconds since 1/1/1970)
	// ... and round up by 1 second since that is the limit of accuracy on the recorded date
	var lastSyncDate = Date.parse(lastSync) * 1000 + 1000000;
	var syncStart = Date.parse(Date()) * 1000;
	var resolveList = [];
	
	LogWrite("Building local dirty list");
	local.BuildDirtyListAS(lastSyncDate, c1);
	
	function c1(localdirty) {
    dirty.local = localdirty;
	
  	LogWrite("Building remote dirty list");
  	remote.BuildDirtyListAS(lastSyncDate, c2);
  }
  
  function c2(remotedirty) {
    dirty.remote = remotedirty;
	
  	LogWrite("Finished building dirty lists");

  	if (local.dirtyList.length > 0 || remote.dirtyList.length > 0) {

  		local.ds.beginUpdateBatch();

  		remote.ProcessDirtyList(local, lastSyncDate, resolveList);
  		local.ProcessDirtyList(remote, lastSyncDate, resolveList);
  		local.FixDanglingReferences(remote, c4);
  	} else {
  		callback(1);
  	}
  }
  
  function c3() {
 		callback(0);
  }
  
  function c4() {
  	remote.FixDanglingReferences(local, c5);
  }

  function c5() {
		local.ds.endUpdateBatch();

		if (resolveList.length > 0) {
			callback(Resolve(local, remote, resolveList));
			return;
		}

    // make sure that any dates that need to be close to our
    // sync finish time get set now
	  remote.RedirtyResourcesAS(gSettings.lastSyncDate, c3);
  }
}

// returns 0 for success, -1 for unresolved conflict

function Resolve(local, remote, resolveList)
{
	var rv = 0;
	const typeMicsum = "http://home.netscape.com/NC-rdf#MicsumBookmark";
	const typeBookmark = "http://home.netscape.com/NC-rdf#Bookmark";
	
	LogWrite("Resolving " + resolveList.length + " conflicts");

	while (resolveList.length > 0) {
		var resource = resolveList.shift();
		
		LogWrite("Resolving " + local.ResourceName(resource));
		
		var resourceTypeLocal = local.GetType(resource);
		var resourceTypeRemote = remote.GetType(resource);

		if (!resourceTypeLocal) {
			alert("ResourceTypeLocal not found for " + resource.Value);
			rv = -1;
			continue;
		}
		
		if (!resourceTypeRemote) {
			alert("ResourceTypeSourceRemote not found for " + resource.Value);
			rv = -1;
			continue;
		}

		if (resourceTypeLocal != resourceTypeRemote) {
      // bookmarks can mutate into micsums and vice versa
      if ((resourceTypeLocal == typeMicsum && resourceTypeRemote == typeBookmark) ||
         (resourceTypeRemote == typeMicsum && resourceTypeLocal  == typeBookmark)) {
         /* mutation has happened */
      } else {

  			alert("Error: Conflicting types for " + resource.Value + "\nLocal: " + resourceTypeLocal + "\nRemote: " + resourceTypeRemote);
  			rv = -1;
  			continue;
      }

		}

		switch (resourceTypeLocal) {
		
			case "http://home.netscape.com/NC-rdf#Folder":
				if (ResolveFolderConflict(local, remote, resource) < 0)
					return -1;
				continue;
				
			case typeBookmark:
			case typeMicsum:
				if (ResolveBookmarkConflict(local, remote, resource, resourceTypeLocal == typeMicsum, resourceTypeRemote == typeMicsum) < 0)
					return -1;
				continue;
				
			case "http://home.netscape.com/NC-rdf#Livemark":
				continue; // skip warnings on LiveMarks
							// XXX note that if some attribute of Livemarks has changed,
							// this code will ignore it, and it will not synch
							
			default:
				break;
		}
		

		if (resource instanceof Components.interfaces.nsIRDFResource) {
			alert ("Resolution conflict with resource " + resource.Value +"\nType is " + resourceTypeLocal);
			rv = -1;
		}
	}
	
	return rv;
}

function InsertOrAppend(container, child, index)
{
  if (index < 1)
    index = 1;
    
  if (index <= container.GetCount())
    container.InsertElementAt(child, index, true);
  else
    container.AppendElement(child);
}

function GetChildByIndex(container, index)
{
  var children = container.GetElements();
  var child = null;

  // bookmark manager doesn't always clean up nextVal properly,
  // leading us to believe that there are more children than there
  // actually are. Check hasMoreElements before calling getNext
  // to avoid crapping out, and return null if we've run out children
  // prematurely.
  for (var i = 1; i <= index; ++i)
    if (children.hasMoreElements())
      child = children.getNext();
    else
      return null;
    
  return child;
}

function IndexInFolder(ds, folder, child)
{
	var container = Components.classes["@mozilla.org/rdf/container;1"].
			createInstance(Components.interfaces.nsIRDFContainer);
			
	container.Init(ds, folder);
	return container.IndexOf(child);
}
  
     
// deal with elements in a folder that appear in local
// but not remote
function ProcessExtras(local, remote, resource, swap)
{
	// process resources in local that are not in remote
	var inval = {localname: "", localdesc: "", remotename: "", remotedesc: "", itemname: ""};
	var retval = {selection: "", canceled: false, source: null, dest: null};
	var container = Components.classes["@mozilla.org/rdf/container;1"].
			createInstance(Components.interfaces.nsIRDFContainer);
	
	// resolve conflict (if any) in direct children
	local.container.Init(local.ds, resource);
	remote.container.Init(remote.ds, resource);
	
	function MoveItemOnRemote() {
	  LogWrite("ProcessExtras: Keeping local for " + local.ResourceName(child));
	  LogWrite("ProcessExtras: Prepping container for " + remote.ResourceName(remoteresource));
	  try {
		  container.Init(remote.ds, remoteresource);
			container.RemoveElement(child, true);
      remote.MarkDirty(remoteresource);
    } catch(e) {
      LogWrite("WARNING: Couldn't init remote folder " + remote.ResourceName(remoteresource));
    }
		InsertOrAppend(remote.container, child, index);
		remote.MarkDirty(resource);
  }
  
  function MoveItemOnLocal() {
	  LogWrite("ProcessExtras: Keeping remote for " + local.ResourceName(child));
	  LogWrite("ProcessExtras: Initing container for " + local.ResourceName(remoteresource));
	  try {
      container.Init(local.ds, remoteresource);
  		InsertOrAppend(container, child, IndexInFolder(remote.ds, remoteresource, child));
    	local.MarkDirty(remoteresource);
    } catch(e) {
      LogWrite("WARNING: Couldn't init local folder " + remote.ResourceName(remoteresource));
    }

		local.container.RemoveElement(child, true);
		index = index - 1;
		local.MarkDirty(resource);
  }


	for (var index = 1; index <= local.container.GetCount(); index = index + 1) {
		var child = GetChildByIndex(local.container, index);
		
		if (!child) {
		  LogWrite("ProcessExtras: Skipping null child");
		  continue;
		}
		
		// XXX if it's a separator, we don't bother to try to fix it up
		// on the remote side, as it's probably a problem caused by
		// ephemeral id's on separators. Unfortunately, this means that
		// we will fail to synch separators created in conflicted folders.
		// Bummer.
		if (child instanceof Components.interfaces.nsIRDFResource &&
				remote.container.IndexOf(child) < 0 &&
				!local.IsSeparator(child)) {
			
			// if doesn't exist at all on remote, kill it here
			if (!remote.ResourceExists(child)) {
			  LogWrite("ProcessExtras: Deleting " + local.ResourceName(child));
				local.RemoveResource(child);
				local.container.RemoveElement(child, true);
				index = index - 1;
				local.MarkDirty(resource);
			}
			
			// if it exists, but has no parent, insert it in remote parent
			else if (!remote.GetParent(child)) {
			  LogWrite("ProcessExtras: Inserting " + local.ResourceName(child) + " into remote");
			  InsertOrAppend(remote.container, child, index);
				remote.MarkDirty(resource);
			}
			
			// if exists at some other parent, we've got to ask the user to decide
			else {
				var remoteresource = remote.GetParent(child);
				
				if (!remote.ResourceExists(remoteresource) || !local.ResourceExists(remoteresource)) {
          MoveItemOnRemote();
          continue;
        }

				const propName = "http://home.netscape.com/NC-rdf#Name";
				const propDesc = "http://home.netscape.com/NC-rdf#Description";

				inval.localname = local.PropString(resource, propName);
				inval.localdesc = local.PropString(resource, propDesc);				
				inval.remotename = remote.PropString(remoteresource, propName);
				inval.remotedesc = remote.PropString(remoteresource, propDesc);
				inval.itemname = local.PropString(child, propName);
				
				// craziness: if swap is set, we're actually processing the remote - local case,
				// not the local - remote case. Swap local/remote variables so the user is presented
				// with a consistent model
				// have to do this on the way out, too.

				if (swap) {
				  var temp;
				  temp = inval.localname;
				  inval.localname = inval.remotename;
				  inval.remotename = temp;
				  
				  temp = inval.localdesc;
				  inval.localdesc = inval.remotedesc;
				  inval.remotedesc = temp;
				}
			
				OpenConflictDialog("chrome://foxmarks/content/foxmarks-parentconflict.xul", 
						retval, inval);
						
				if (retval.canceled)
					return -1;
	
				if (swap) {
					retval.selection = (retval.selection == "local" ? "remote" : "local");
				}
				
				// user wants to keep local, so move remote
				if (retval.selection == "local") {
          MoveItemOnRemote();
				} else {	// keep remote
          MoveItemOnLocal();
				}
			}
		} else {
		  LogWrite("ProcessExtras: Skipping " + local.ResourceName(child));
		}
	}
	
	return 0;
}

function SameOrder(local, remote, resource)
{
	local.container.Init(local.ds, resource);
	remote.container.Init(remote.ds, resource);

	var localChildren = local.container.GetElements();
	var remoteChildren = remote.container.GetElements();
	var localResource;
	var remoteResource;
	
	while (localChildren.hasMoreElements() && remoteChildren.hasMoreElements()) {
		localResource = localChildren.getNext();
		remoteResource = remoteChildren.getNext();
		if (localResource != remoteResource &&
				!(local.IsSeparator(localResource) && remote.IsSeparator(remoteResource))) {
			return false;
		}
	}

	// we tried using GetCount() to determine number of children, but
	// GetCount returns an incorrect number of children on sequences that
	// have been modified by the user in the BookmarksManager view.
	// Instead, we just do a sanity check here to make sure we've exhausted
	// both enumerators.
	if (localChildren.hasMoreElements() != remoteChildren.hasMoreElements()) {
		dump("Error: Different number of children in " + resource.Value);
		return false;
	}
	
	return true;
}
	
// return values:
// -1: canceled by user

function ResolveFolderConflict(local, remote, resource)
{
	var retval = {selection: "", canceled: false, source: null, dest: null};
	var inval = {localname: "", localdesc: "", remotename: "", remotedesc: "", itemname: ""};

	inval.local = local;
	inval.remote = remote;
	
	const propName = "http://home.netscape.com/NC-rdf#Name";
	const propDesc = "http://home.netscape.com/NC-rdf#Description";


	// gather up some values we'll need
	inval.localname = local.PropString(resource, propName);
	inval.remotename = remote.PropString(resource, propName);
	inval.localdesc = local.PropString(resource, propDesc);
	inval.remotedesc = remote.PropString(resource, propDesc);
	
	// resolve conflict (if any) in direct folder properties
	if (inval.localname != inval.remotename || inval.localdesc != inval.remotedesc) {
		OpenConflictDialog("chrome://foxmarks/content/foxmarks-folderconflict.xul", 
				retval, inval);
		if (retval.canceled) {
			return -1;
		}
		
		retval.dest.CopyProperty(resource, propName, retval.source);
		retval.dest.CopyProperty(resource, propDesc, retval.source);
		retval.dest.MarkDirty(resource);
	}
	
	// now deal with children that appear on one side,
	// but not the other
	LogWrite("Calling ProcessExtras (local, remote) for " + local.ResourceName(resource));
	if (ProcessExtras(local, remote, resource, false) < 0) {
		return -1;
	}

	LogWrite("Calling ProcessExtras (remote, local) for " + local.ResourceName(resource));	
	if (ProcessExtras(remote, local, resource, true) < 0) {
		return -1;
	}
	
  if (ResolveSeparatorConflict(local, remote, resource) < 0) {
    return -1;
  }
		
	// now check ordering and ask user to select local or remote
	// ordering if they are different
	if (!SameOrder(local, remote, resource)) {
		inval.itemname = local.PropString(resource, propName);
		inval.localdatasource = local.ds;
		inval.remotedatasource = remote.ds;
		inval.localroot = resource.Value;
		inval.remoteroot = resource.Value;
		
		OpenConflictDialog("chrome://foxmarks/content/foxmarks-orderconflict.xul",
		  retval, inval);
			
		if (retval.canceled) {
			return -1;
		}
		
		retval.dest.CopyChildren(resource, retval.source);
		retval.dest.MarkDirty(resource);
		retval.dest.FixSeparators(resource);		// we may have introduced more orphaned separators, so fix them here
	}
	
	return 0;
}
	
function ResolveBookmarkConflict(local, remote, resource, localIsMicsum, remoteIsMicsum)
{
	var retval = {selection: "", canceled: false, source: null, dest: null};
	var inval = {localname: "", localdesc: "", localurl: "", localkeyword: "", localwebpanel: false,
							remotename: "", remotedesc: "", remoteurl: "", remotekeyword: "", remotewebpanel: false};

	inval.local = local;
	inval.remote = remote;
	
	const propName = "http://home.netscape.com/NC-rdf#Name";
	const propGeneratedTitle = "http://home.netscape.com/NC-rdf#GeneratedTitle";
	const propGenURI = "http://home.netscape.com/NC-rdf#GenURI";
	const propDesc = "http://home.netscape.com/NC-rdf#Description";
	const propURL = "http://home.netscape.com/NC-rdf#URL";
	const propKeyword = "http://home.netscape.com/NC-rdf#ShortcutURL";
	const propWebPanel = "http://home.netscape.com/NC-rdf#WebPanel";
	const propType = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type";


	// gather up some values we'll need
	inval.localname = local.PropString(resource, localIsMicsum ? propGeneratedTitle : propName);
	inval.remotename = remote.PropString(resource, remoteIsMicsum ? propGeneratedTitle : propName);
	inval.localurl = local.PropString(resource, propURL);
	inval.remoteurl = remote.PropString(resource, propURL);
	inval.localkeyword = local.PropString(resource, propKeyword);
	inval.remotekeyword = remote.PropString(resource, propKeyword);
	inval.localwebpanel = local.PropString(resource, propWebPanel);
	inval.remotewebpanel = remote.PropString(resource, propWebPanel);
	inval.localdesc = local.PropString(resource, propDesc);
	inval.remotedesc = remote.PropString(resource, propDesc);
	
	// resolve conflict (if any) in direct bookmark properties
	if (inval.localname != inval.remotename || 
		inval.localurl != inval.remoteurl ||
		inval.localkeyword != inval.remotekeyword ||
		inval.localwebpanel != inval.remotewebpanel ||
		inval.localdesc != inval.remotedesc) {
		OpenConflictDialog("chrome://foxmarks/content/foxmarks-bookmarkconflict.xul",
				retval, inval);
		if (retval.canceled) {
			return -1;
		}
		
		var d = retval.dest;
		var s = retval.source;

    // copy individual properties (if they exist)
		d.CopyProperty(resource, propName, s);
		d.CopyProperty(resource, propDesc, s);
		d.CopyProperty(resource, propURL, s);
		d.CopyProperty(resource, propKeyword, s);
		d.CopyProperty(resource, propWebPanel, s);
		d.CopyProperty(resource, propType, s);
		d.CopyProperty(resource, propGenURI, s);
		retval.dest.MarkDirty(resource);
	}
	
	return 0;
}

function OpenConflictDialog(url, retval, inval)
{
  // find a browser window to use as a parent for openDialog
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService();
  var windowManagerInterface = windowManager.QueryInterface(Components.interfaces.nsIWindowMediator);
  var topWindow = windowManagerInterface.getMostRecentWindow("foxmarks:progress");

  // if we can't find one of our own windows, try any window  
  if (!topWindow)
    topWindow = windowManagerInterface.getMostRecentWindow(null);
    
  // We couldn't find a window, so pretend user canceled
  if (!topWindow) {
    retval.canceled = true;
    return;
  }
  
  topWindow.openDialog(url, "_blank", "chrome,dialog,modal,centerscreen", retval, inval);
}

function ResolveSeparatorConflict(local, remote, resource)
{
  const propName = "http://home.netscape.com/NC-rdf#Name";
  var localseps = FindSeparators(local, resource);
  var remoteseps = FindSeparators(remote, resource);
  var localmatches = [];
  var remotematches = [];
  resource.QueryInterface(Components.interfaces.nsIRDFResource);
  
  LogWrite("Entered ResolveSeparatorConflict (local,remote) is (" + localseps.length + "," + remoteseps.length + ")");

  // match 'em by id
  MatchSep(function(x, y) { return (x.id == y.id); });
  
  // match 'em by (non-null) name
  MatchSep(function(x, y) { return (x.name.length && x.name == y.name); } );

  // match 'em by position
  for (var t = 0; localseps.length > 0 && remoteseps.length > 0; t += 1) {
    MatchSep(function(x, y) { return (Math.abs(x.pos - y.pos) <= t); } );
  }
  
  if (localseps.length > 0 || remoteseps.length > 0) {
    // put up the conflict dialog
  	var retval = {selection: "", canceled: false, source: null, dest: null};
  	var inval = {};

    inval.local = {seps: localseps, set:local, matches: localmatches};
    inval.remote = {seps: remoteseps, set: remote, matches: remotematches};
		inval.itemname = local.PropString(resource, propName);

		inval.localdatasource = local.ds;
		inval.remotedatasource = remote.ds;
		inval.localroot = resource.Value;
		inval.remoteroot = resource.Value;
		
    OpenConflictDialog("chrome://foxmarks/content/foxmarks-separatorconflict.xul",
      retval, inval);
    if (retval.canceled)
      return -1;

    // set up a container on dest, where we'll be making changes
    retval.dest.set.container.Init(retval.dest.set.ds, resource);
    
    // are we inserting or deleting?
    if (retval.source.seps.length) {     // inserting
      while (retval.source.seps.length) {
        var sep = retval.source.seps.pop();
        InsertOrAppend(retval.dest.set.container, sep.id, sep.pos + 1);
        retval.dest.set.CopyResource(retval.source.set, sep.id, false);
        LogWrite("Inserted separator at index " + sep.pos + 1);
      }
    } else if (retval.dest.seps.length) { // deleting
      while (retval.dest.seps.length) {
        var sep = retval.dest.seps.pop();
        retval.dest.set.container.RemoveElement(sep.id, true);
      }
    }
    
    // for each matched separator, copy its name (deleting if necessary)
    var predName = retval.dest.set.rdfs.GetResource(propName);
    while (retval.dest.matches.length > 0) {
      var d = retval.dest.matches.pop();
      var s = retval.source.matches.pop();
      if (s.name != d.name) {
        // remove existing old prop if it exists
        var oldname = retval.dest.set.GetTarget(d.id, predName, true);
        if (oldname) retval.dest.set.Unassert(d.id, predName, oldname, true);
        retval.dest.set.Assert(d.id, predName, retval.dest.set.GetResource(s.name), true);
      }
    }
  } else {
    // we're not going to display a conflict dialog, but there's one
    // special case we'd still like to handle: user dirties a sep (by
    // changing its name) in a conflicted folder
    while (localmatches.length) {
      var l = localmatches.pop();
      var r = remotematches.pop();
      
      if (l.id != r.id) {
        if (local.ResourceExists(r.id) && !remote.ResourceExists(l.id))
          CopyName(local, l.id, local, r.id);
        else if (remote.ResourceExists(l.id) && !local.ResourceExists(r.id))
          CopyName(remote, r.id, remote, l.id);
      }
    }
  }
  
  return 0;

  function CopyName(dest, destres, src, srcres) {
    var predName = dest.rdfs.GetResource(propName);
    var oldname = dest.ds.GetTarget(destres, predName, true);
    if (oldname) dest.ds.Unassert(destres, predName, oldname, true);
    var newname = src.ds.GetTarget(srcres, predName, true);
    if (newname) dest.ds.Assert(destres, predName, newname, true);
  }

  function FindSeparators(set, resource) {
  
    var seps = [];
    set.container.Init(set.ds, resource);
    var children = set.container.GetElements();

    for (var index = 0; children.hasMoreElements(); index += 1) {
      var child = children.getNext();
      
      if (set.IsSeparator(child)) {
        LogWrite("Found separator " + set.ResourceName(child));
        var s = new Object();
        s.id = child;
        s.pos = index;
        s.name = set.PropString(child, propName);
        seps.push(s);
      }
    }
    return seps;
  }
  
  function MatchSep(f) {
    for (var i in localseps) {
      for (var j in remoteseps) {
        if (f(localseps[i], remoteseps[j])) {
          localmatches.push(localseps[i]);
          remotematches.push(remoteseps[j]);
          localseps.splice(i, 1);
          remoteseps.splice(j, 1);
          break;
        }
      }
    }
  }
}
